/**
 * Date:            Time:       User:               Description:
 * 14.03.2006       20:20       Smart Max           creation of the file IWritablePatriciaTrie.java
 *
 */
package org.max.trie;

import java.util.List;

/**
 * IWritablePatriciaTrie
 *
 * @author  Maxim Podkolzine
 */
public interface IWritablePatriciaTrie extends IPatriciaTrie {
    /**
     * Adds a new string to the trie.
     * <p>
     * Method consecutively traverse the trie edges, which coincide with prefix of the key. When bifurcation
     * occurs, either adding a single new node can happen or adding an edge to an existing node (it is possible
     * also that both actions happen). In any case, at most two new nodes will be added to the trie independently
     * from length of string and size of trie.
     * <p>
     * Method works only for modifiable Patricia tries.
     *
     * @param   path                                the string to be added
     * @return                                      the node corresponding to the <tt>path</tt> in the trie
     * @see     #get(java.util.List<org.max.trie.ISymbol>)
     * @see     #isModifiable()
     */
    public IWritableNode add(List<ISymbol> path);

    /**
     * @see     #add(java.util.List<org.max.trie.ISymbol>)
     */
    public INode get(List<ISymbol> string);

    /**
     */
    public INode getRoot();

    /**
     */
    public int size();

    /**
     * Returns whether the trie is modifiable.
     * <p>
     * For newly constructed tries this value always return <tt>true</tt>. But if the trie was serialized
     * (see {@link org.max.trie.impl.PatriciaTrieSerializer}), the uploaded trie can not perform adding.
     * This is caused by loss of map of symbols on edges (see Patricia trie structure in {@link IPatriciaTrie}).
     * In fact, if there exists correct map of symbols on edges, the trie can be simply transformed to modifiable
     * one by method {@link #setModifiable(boolean)}. Otherwise further addings will work incorrectly.
     *
     * @return                                      true iff the trie is modifiable
     * @see     #setModifiable(boolean)
     */
    public boolean isModifiable();

    /**
     * Sets the modification value of the trie.
     * <p>
     * Note, that to make the trie modifiable, the map of symbols on edges must be setted. It can exist in a trie
     * since creating (if it was not released, see {@link #release()}), or can be uploaded from file (or any input
     * stream), if the map was serialized in advance (see {@link org.max.trie.impl.PatriciaTrieSerializer}).
     * Otherwise further addings to the trie will work incorrectly.
     *
     * @param   modifiable                          the new modification value
     * @see     #isModifiable()
     * @see     #release()
     */
    public void setModifiable(boolean modifiable);

    /**
     * Releases the map of symbols on edges of the trie.
     * <p>
     * This map is needed only for additions to the trie and is not needed in searching (see Patricia trie
     * structure in {@link IPatriciaTrie}). This means that after trie constructing it is possible that
     * the trie will be searched in only. In this case, the memory occupied by map can be released using
     * this method. Before this, it can be serialized by {@link org.max.trie.impl.PatriciaTrieSerializer}.
     */
    public void release();
}
