/**
 * Date:            Time:       User:               Description:
 * 14.03.2006       20:20       Smart Max           creation of the file Node.java
 * 26.03.2006       21:26       Smart Max           references added
 *
 */
package org.max.trie.impl;

import static org.max.NativeSystem.*;

import org.max.trie.IWritableNode;
import org.max.trie.ISymbol;
import org.max.trie.IReference;

import java.util.*;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.IOException;

/**
 * Node
 *
 * @author  Maxim Podkolzine
 */
class Node implements IWritableNode, Serializable {
    private static final long serialVersionUID = 100036478973486L;

    // Data

    /**
     * Holds the list of references (i.e. the objects which have some additional
     * information about a node).
     * <p>
     * Only marked nodes have references, for others this field is <tt>null</tt>.
     */
    private List<IReference> myReferences;

    /**
     * Holds the depth from root (in symbols)
     */
    private int myDepth;

    /**
     * Holds the reference to the parent node.
     * <p>
     * The field is transient to serialize the node separately from parent.
     */
    private transient IWritableNode myParent;

    /**
     * Holds the incoming symbol.
     */
    private ISymbol myIncomingSymbol;

    /**
     * Holds the map of children.
     * <p>
     * The values can be instance of INode, or instance of String
     * (in this case it is a far link to the node in another block).
     * <p>
     * The field is transient to serialize the node separately from child nodes.
     */
    private transient Map<ISymbol, Object> myChildren;




    // Constructor

    /**
     * Creates a node of specified depth, parent and incoming symbol
     *
     * @param   depth                               the depth
     * @param   parent                              the parent node
     * @param   incomingSymbol                      the symbol on incoming edge
     */
    public Node(int depth, IWritableNode parent, ISymbol incomingSymbol) {
        myDepth = depth;
        myParent = parent;
        myIncomingSymbol = incomingSymbol;
        myChildren = new HashMap<ISymbol, Object>();
    }

    /**
     * Creates a node with no parent and incoming symbol and the depth value equal to <tt>0</tt>.
     * <p>
     * In other words, this constructor creates a root node.
     */
    public Node() {
        this(0, null, null);
    }




    // Main interface

    /**
     * Creates a children map.
     * <p>
     * The method is used to initialize a map of children after deserialization
     * (field <tt>myChildren</tt> is transient).
     *
     * @see     org.max.trie.impl.Block#getSubTrie()
     */
    public void createChildrenMap() {
        myChildren = new HashMap<ISymbol, Object>();
    }

    /**
     * Returns a string representation of the node, which consists of depth value and number of children
     *
     * @return                                      a string representation of node
     */
    public String toString() {
        return "Depth = " + myDepth + "\n" +
                "Children number = " + myChildren.size();
    }

    /**
     * Returns unmodifiable list of references of the node, or <tt>null</tt>, if it has not references
     *
     * @return                                      the list of references
     */
    public List<IReference> getReferences() {
        if (myReferences == null) {
            return null;
        } else {
            return Collections.unmodifiableList(myReferences);
        }
    }

    /**
     * Adds new reference to the node.
     * If it will be the first reference, then the list <tt>myReferences</tt> will be created and
     * a specified reference will be added.
     *
     * @param   reference                           the reference to add
     */
    public void addReference(IReference reference) {
        if (myReferences == null) {
            myReferences = new ArrayList<IReference>(1);
        }
        myReferences.add(reference);
    }

    /**
     * Returns the depth of the node
     *
     * @return                                      the depth of the node
     */
    public int getDepth() {
        return myDepth;
    }

    /**
     * Updates the depth of a node
     *
     * @param   depth                               the new depth
     */
    public void setDepth(int depth) {
        myDepth = depth;
    }

    /**
     * Returns the parent of the node
     *
     * @return                                      the parent node
     */
    public IWritableNode getParent() {
        return myParent;
    }

    /**
     * Updates the parent of the node
     *
     * @param   parent                              the new parent
     */
    public void setParent(IWritableNode parent) {
        myParent = parent;
    }

    /**
     * Returns the incoming symbol of the node
     *
     * @return                                      the incoming symbol
     */
    public ISymbol getIncomingSymbol() {
        return myIncomingSymbol;
    }

    /**
     * Updates the incoming symbol of the node
     *
     * @param   incomingSymbol                      the new incoming symbol
     */
    public void setIncomingSymbol(ISymbol incomingSymbol) {
        myIncomingSymbol = incomingSymbol;
    }

    /**
     * Returns the unmodifiable set of symbols on outcoming edges
     *
     * @return                                      the set of outcoming symbols
     */
    public Set<ISymbol> getOutcomingSymbols() {
        return Collections.unmodifiableSet(myChildren.keySet());
    }

    /**
     * Returns wheter the node has specified label among outcoming symbols
     *
     * @param   label                               the label on edge
     * @return                                      true iff there is an edge with such label on it
     */
    public boolean hasLabel(ISymbol label) {
        return myChildren.containsKey(label);
    }

    /**
     * Adds the child to the node.
     * <p>
     * If <tt>child == null</tt>, method throws <tt>NullPointerException</tt>.
     * Method does not update the parent of child. It only registers a new child of this node, using
     * the incoming symbol of child. If such label already exists among outcoming symbols of the node,
     * the <tt>LabelDuplicatedException</tt> exception is thrown.
     *
     * @param   child                               new child
     * @throws  NullPointerException                if <tt>child</tt> is <tt>null</tt>
     * @throws  LabelDuplicatedException            if the node already has the label
     *                                              (incoming symbol of <tt>child</tt>)
     */
    public void addChild(IWritableNode child) {
        if (child == null) {
            throw new NullPointerException("The child is null");
        }

        ISymbol label = child.getIncomingSymbol();

        if (hasLabel(label)) {
            throw new LabelDuplicatedException("The node already has the label = " + label);
        }

        myChildren.put(label, child);
    }

    /**
     * Adds the far child to the node.
     * <p>
     * If specified label already exists among outcoming symbols of the node,
     * the <tt>LabelDuplicatedException</tt> exception is thrown.
     *
     * @param   label                               the label
     * @param   blockName                           the block name
     * @throws  LabelDuplicatedException            if the node already has the <tt>label</tt>
     */
    public void addChild(ISymbol label, String blockName) {
        if (hasLabel(label)) {
            throw new LabelDuplicatedException("The node already has the label = " + label);
        }

        myChildren.put(label, blockName);
    }

    /**
     * Returns the child of the node.
     * <p>
     * Method gets the close child if there is direct reference on this object;
     * if children map has far reference on child (the block name), corresponding block
     * is uploaded and the result is root of sub-trie inside this block;
     * if children map does not contain the label, method returns <tt>null</tt>.
     * <p>
     * At worst method requires to upload the far block and then construct the trie from it
     * ({@link Block#getSubTrie()}, which can take much time. The average time
     * of work of this method is good enough.
     *
     * @param   label                               the label of the edge to requested child
     * @return                                      the child node
     */
    public IWritableNode getChild(ISymbol label) {
        Object obj = myChildren.get(label);

        if (obj == null) {
            return null;
        } else if (obj instanceof IWritableNode) {
            return (IWritableNode) obj;
        } else  if (obj instanceof String) {
            try {
                ObjectInputStream input = new ObjectInputStream(newInputStream((String) obj));
                Block block = (Block) input.readObject();
                                                                  // Here the parent of sub-trie root node
                return block.getSubTrie();                        // can be setted. It does not needed while
                                                                  // blocks are independent from each other
            } catch (IOException e) {
                System.out.println("I/O error occurs while block reading.\n" + e);
                return null;
            } catch (ClassNotFoundException e) {
                System.out.println("Cannot read block.\n" + e);
                return null;
            }
        } else {
            throw new RuntimeException("Internal error. Node structure is incorrect. child = " + obj);  
        }
    }

    /**
     * Changes the child which corresponds to specified label.
     * <p>
     * If the <tt>label</tt> is not recognized among outcoming edges of the node, the exception
     * <tt>LabelNotFoundException</tt> raises.
     *
     * @param   label                               the symbol
     * @param   node                                the new child node
     * @throws  LabelNotFoundException              if specified label was not found
     */
    public void changeChild(ISymbol label, IWritableNode node) {
        if (!hasLabel(label)) {
            throw new LabelNotFoundException("Symbol not found. label = " + label);
        }

        myChildren.put(label, node);
    }

    /**
     * Changes the far child which corresponds to specified label.
     * <p>
     * If the <tt>label</tt> is not recognized among outcoming edges of the node, the exception
     * <tt>LabelNotFoundException</tt> raises.
     *
     * @param   label                               the symbol
     * @param   blockName                           the block name
     * @throws  LabelNotFoundException              if specified label was not found
     */
    public void changeChild(ISymbol label, String blockName) {
        if (!hasLabel(label)) {
            throw new LabelNotFoundException("Symbol not found. label = " + label);
        }

        myChildren.put(label, blockName);
    }
}
