/**
 * Date:            Time:       User:               Description:
 * 01.05.2006       12:57       Smart Max           creation of the file PatriciaTrieDeserializer.java
 *
 */
package org.max.trie.impl;

import static org.max.NativeSystem.*;

import org.max.trie.IPatriciaTrie;
import org.max.trie.IWritablePatriciaTrie;
import org.max.trie.IWritableNode;
import org.max.trie.ISymbol;

import java.io.*;
import java.util.Map;
import java.util.List;

/**
 * PatriciaTrieDeserializer
 *
 * @author  Maxim Podkolzine
 */
public class PatriciaTrieDeserializer {
    /**
     * Reads the root block from specified directory.
     * <p>
     * Method first tries to read first string from file "meta.dat" and after this to upload the
     * block by read name. By the contract, the block is root block of the trie.
     *
     * @param   directory                           the source directory
     * @return                                      the root block of the trie
     * @throws  IOException                         if I/O errors occur
     * @throws  ClassNotFoundException              if class of a serialized block cannot be found
     */
    private Block readRootBlock(String directory) throws IOException, ClassNotFoundException {
        BufferedReader reader = new BufferedReader(newReader(directory + Separator + "meta.dat"));
        String rootBlockName = reader.readLine();
        reader.close();

        ObjectInputStream input = new ObjectInputStream(newInputStream(directory + Separator + rootBlockName));
        Block block = (Block) input.readObject();
        input.close();

        return block;
    }

    /**
     * Reads the unmodifiable trie from specified directory.
     * <p>
     * At first method gets the root block of the trie and after this constructs the trie using
     * the root of the corresponding sub-trie. The map of symbols on edges is not uploaded, thus
     * no further addings to the trie will be admissible.
     *
     * @param   directory                           the source directory
     * @return                                      the Patricia Trie interface
     * @throws  IOException                         if I/O errors occur
     * @throws  ClassNotFoundException              if class of a serialized block cannot be found
     */
    public IPatriciaTrie readTrie(String directory) throws IOException, ClassNotFoundException {
        Block block = readRootBlock(directory);
        return new PatriciaTrie(block.getSubTrie(), null, false);
    }

    /**
     * Reads the modifiable trie from specified directory.
     * <p>
     * At first method gets the root block of the trie, then gets the map of symbols on edges from
     * "triemap.dat" file and after this constructs the trie using the root of the corresponding sub-trie
     * and obtained map.
     *
     * @param   directory                           the source directory
     * @return                                      the Patricia Trie writable interface
     * @throws  IOException                         if I/O errors occur
     * @throws  ClassNotFoundException              if class of a serialized block cannot be found
     */
    public IWritablePatriciaTrie readFullTrie(String directory) throws IOException, ClassNotFoundException {
        Block block = readRootBlock(directory);

        ObjectInputStream input = new ObjectInputStream(newInputStream(directory + Separator + "triemap.dat"));
        Map<IWritableNode, List<ISymbol>> map = (Map<IWritableNode, List<ISymbol>>) input.readObject();
        input.close();

        return new PatriciaTrie(block.getSubTrie(), map, true);
    }
}
