/**
 * Date:            Time:       User:               Description:
 * 01.04.2006       16:27       Smart Max           creation of the file IndexFabric.java
 *
 */
package org.max.xml.index;

import org.max.trie.IPatriciaTrie;
import org.max.trie.ISymbol;
import org.max.util.BidirectionalMap;

/**
 * IndexFabric
 * <p>
 * Consists of Patricia trie and the amount of designators maps.
 *
 * @author  Maxim Podkolzine
 */
public class IndexFabric {
    // Data

    /**
     * Holds the Patricia trie
     */
    private IPatriciaTrie myTrie;

    /**
     * Holds the map of text symbols
     */
    private final BidirectionalMap<String, ISymbol> myTextMap;

    /**
     * Holds the map of symbols corresponding to tags in XML files
     */
    private final BidirectionalMap<String, ISymbol> myTagMap;

    /**
     * Holds the map of symbols corresponding to attributes in XML files
     */
    private final BidirectionalMap<String, ISymbol> myAttrMap;




    // Constructor

    /**
     * Creates the Index Fabric with specified Patricia trie and designators maps
     *
     * @param   trie                                the Patricia trie
     * @param   textMap                             the map of text designators
     * @param   tagMap                              the map of tag (element) designators
     * @param   attrMap                             the map of attribute designators
     */
    public IndexFabric(IPatriciaTrie trie,
                       BidirectionalMap<String, ISymbol> textMap,
                       BidirectionalMap<String, ISymbol> tagMap,
                       BidirectionalMap<String, ISymbol> attrMap) {
        myTrie = trie;
        myTextMap = textMap;
        myTagMap = tagMap;
        myAttrMap = attrMap;
    }




    // Main interface

    /**
     * Returns the Patricia trie of this Index Fabric
     *
     * @return                                      the Patricia trie of the Index Fabric
     */
    public IPatriciaTrie getTrie() {
        return myTrie;
    }

    /**
     * Returns the map of text designators of this Index Fabric
     *
     * @return                                      the map of text designators of the Index Fabric
     */
    public BidirectionalMap<String, ISymbol> getTextMap() {
        return myTextMap;
    }

    /**
     * Returns the map of tag designators of this Index Fabric
     *
     * @return                                      the map of tag designators of the Index Fabric
     */
    public BidirectionalMap<String, ISymbol> getTagMap() {
        return myTagMap;
    }

    /**
     * Returns the map of attribute designators of this Index Fabric
     *
     * @return                                      the map of attribute designators of the Index Fabric
     */
    public BidirectionalMap<String, ISymbol> getAttrMap() {
        return myAttrMap;
    }
}