/**
 * Date:            Time:       User:               Description:
 * 13.03.2006       11:58       Smart Max           creation of the file XMLBuilder.java
 *
 */
package org.max.xml.index;

import static org.w3c.dom.Node.*;
import static org.max.NativeSystem.*;

import org.max.trie.IWritablePatriciaTrie;
import org.max.trie.ISymbol;
import org.max.trie.IWritableNode;
import org.max.trie.impl.PatriciaTrieFactory;
import org.max.util.BidirectionalMap;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;
import java.io.InputStream;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;

import net.sf.xanswer.indexing.IndexException;

/**
 * XMLBuilder.
 * It is Index Fabric Builder from XML.
 * <p>
 * After constructing, the result <tt>IndexFabric</tt> is unmodifiable
 * <p>
 * @author  Maxim Podkolzine
 */
public class XMLBuilder {
    // Data

    /**
     * Holds the constructed Patricia trie
     */
    private IWritablePatriciaTrie myTrie;

    /**
     * Holds the map of text symbols
     */
    private final BidirectionalMap<String, ISymbol> myTextMap;

    /**
     * Holds the map of symbols corresponding to tags in XML files
     */
    private final BidirectionalMap<String, ISymbol> myTagMap;

    /**
     * Holds the map of symbols corresponding to attributes in XML files
     */
    private final BidirectionalMap<String, ISymbol> myAttrMap;

    /**
     * Holds the SAX handler of XML parsing events
     */
    private final Handler myHandler;




    // Constructor

    /**
     * Creates a new instance of <tt>XMLBuilder</tt>.
     * <p>
     * After creation the trie and all maps are empty. The newly created Index Fabric is modifiable.
     */
    public XMLBuilder() {
        myTrie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();
        myTextMap = new BidirectionalMap<String, ISymbol>();
        myTagMap = new BidirectionalMap<String, ISymbol>();
        myAttrMap = new BidirectionalMap<String, ISymbol>();
        myHandler = new Handler();
    }

    /**
     * Creates a new instance of <tt>XMLBuilder</tt> using an existed Index Fabric.
     * <p>
     * After creation builder holds an existed Index Fabric (not necessarily empty). The Index Fabric is
     * casted to modifiable. Thus, if Patricia trie of this Index Fabric does not hold the map of
     * symbols on edges (see {@link org.max.trie.IPatriciaTrie}), new documents addings will cause exception
     * raising.
     *
     * @param   indexFabric                         the Index Fabric
     */
    public XMLBuilder(IndexFabric indexFabric) {
        myTrie = (IWritablePatriciaTrie) indexFabric.getTrie();
        myTrie.setModifiable(true);

        myTextMap = indexFabric.getTextMap();
        myTextMap.setModifiable(true);

        myTagMap = indexFabric.getTagMap();
        myTagMap.setModifiable(true);

        myAttrMap = indexFabric.getAttrMap();
        myAttrMap.setModifiable(true);

        myHandler = new Handler();
    }




    // Main interface

    // Options

    /**
     * Holds the boolean value: it is true iff all whitespaces are ignored.
     * <p>
     * By default, the value is false.
     */
    private static boolean ourIgnoreWhitespace = false;

    /**
     * Returns whether {@link #ourIgnoreWhitespace} option is on
     *
     * @return                                      true iff {@link #ourIgnoreWhitespace} option is on
     */
    public static boolean isIgnoreWhitespace() {
        return ourIgnoreWhitespace;
    }

    /**
     * Updates the {@link #ourIgnoreWhitespace} option
     *
     * @param   ignoreWhitespace                    the new value of blanks (or whitespaces) omitting
     */
    public static void setIgnoreWhitespace(boolean ignoreWhitespace) {
        ourIgnoreWhitespace = ignoreWhitespace;
    }

    /**
     * Holds the boolean value: it is true iff Patricia trie is constructed by all prefixes of
     * root-to-leaf paths in XML. In other words, all paths root-to-node are added.
     * <p>
     * By default, the value if true.
     */
    private static boolean ourAddPathPrefixes = true;

    /**
     * Returns whether {@link #ourAddPathPrefixes} option is on
     *
     * @return                                      true iff {@link #ourAddPathPrefixes} option is on
     */
    public static boolean isAddPathPrefixes() {
        return ourAddPathPrefixes;
    }

    /**
     * Updates the {@link #ourAddPathPrefixes} option
     *
     * @param   addPathPrefixes                     the new value of path prefixes adding
     */
    public static void setAddPathPrefixes(boolean addPathPrefixes) {
        ourAddPathPrefixes = addPathPrefixes;
    }

    /**
     * Restores the default options: sets off the {@link #ourIgnoreWhitespace} option and
     * sets the {@link #ourAddPathPrefixes} option on.
     */
    public static void restoreDefaults() {
        ourIgnoreWhitespace = false;
        ourAddPathPrefixes = true;
    }

    // Add XML method

    /**
     * Returns the symbol of certain type that corresponds to specified name. At first, the name is sought in
     * map and corresponding symbol is returned if it was found. Else method constructs the new
     * symbol, which is not contained in map before, and returns it.
     * <p>
     * To construct new symbol different subsequences of <tt>name</tt> are tested in the following order:
     * <ul>
     * <li> all one-symbol substrings
     * <li> the first symbol concatenated with all other one-symbol substrings
     * <li> first two symbols and all one-symbol substrings, and so on
     * </ul>
     * They all are tested (i.e. checked if correponding symbol of type <tt>type</tt> is free). The first
     * proper symbol is added to the map and returned.
     * <p>
     * If all concerned subsequences are used for other names, the <tt>name</tt> concatenated with
     * 0, 1, 2, ... are tested until success.
     *
     * @param   name                                the name
     * @param   map                                 the map to search in
     * @param   type                                the type of symbol
     * @return                                      the correponding symbol of certain type
     */
    private ISymbol makeSymbol(String name, BidirectionalMap<String, ISymbol> map, int type) {
        if (map.containsA(name)) {
            return map.getB(name);
        }

        for (int i = 0; i < name.length(); i++) {
            StringBuilder builder = new StringBuilder(name.substring(0, i));
            for (int j = i; j < name.length(); j++) {
                builder.append(name.charAt(j));
                ISymbol symbol = new Symbol(builder.toString(), type);

                if (!map.containsB(symbol)) {
                    map.put(name, symbol);
                    return symbol;
                }

                builder.setLength(i);
            }
        }

        int i = 0;
        StringBuilder builder = new StringBuilder(name);
        while (true) {
            builder.append(i);
            ISymbol symbol = new Symbol(builder.toString(), type);

            if (!map.containsB(symbol)) {
                map.put(name, symbol);
                return symbol;
            }

            builder.setLength(name.length());
            i++;
        }
    }

    /**
     * Returns the list of text symbols which corresponds to specified text.
     * <p>
     * The new characters are added to the <tt>myTextMap</tt> map, others are simply taken from
     * this map. After this, they all are added into one list in right order.
     *
     * @param   text                                the text
     * @return                                      the list of symbols which corresponds to specified text
     */
    private List<ISymbol> makeTextSymbols(String text) {
        List<ISymbol> result = new ArrayList<ISymbol>();

        for (int i = 0; i < text.length(); i++) {
            String c = String.valueOf(text.charAt(i));
            ISymbol symbol = myTextMap.getB(c);

            if (symbol == null) {
                symbol = new Symbol(c, TEXT_NODE);
                myTextMap.put(c, symbol);
            }

            result.add(symbol);
        }

        return result;
    }

    /**
     * Inner class which proceeds the SAX events, such as
     * <ul>
     * <li> start of an element (tag); method {@link Handler#startElement(String, String, String, Attributes)}
     *      handles this event
     * <li> end of an element (tag); method {@link Handler#endElement(String, String, String)}
     *      handles this event
     * <li> notification of character data inside an element; method {@link Handler#characters(char[], int, int)}
     *      handles this event
     * </ul>
     *
     * Class holds the current path, which is added to the trie when needed.
     */
    private class Handler extends DefaultHandler {
        // Data

        /**
         * Holds the index of current tag
         */
        private int tagIndex;

        /**
        * Holds current document ID (for example, the name of XML file)
        */
        private String docID;

        /**
         * Holds the current path
         */
        private List<ISymbol> path;

        // Main interface

        /**
         * Restarts the handler.
         * <p>
         * Method sets new document identifier, nulls the tag index and path.
         *
         * @param   docID                           the document ID
         */
        public void restart(String docID) {
            this.docID = docID;
            tagIndex = 0;
            path = new ArrayList<ISymbol>();
        }

        /**
         *
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
            tagIndex++;
            path.add(makeSymbol(qName, myTagMap, ELEMENT_NODE));
            if (ourAddPathPrefixes) {
                myTrie.add(path);
            }

            for (int i = 0; i < attributes.getLength(); i++) {
                int size = path.size();
                path.add(makeSymbol(attributes.getQName(i), myAttrMap, ATTRIBUTE_NODE));
                if (ourAddPathPrefixes) {
                    myTrie.add(path);
                }

                path.addAll(makeTextSymbols(attributes.getValue(i)));
                IWritableNode node = myTrie.add(path);
                node.addReference(new XMLReference(tagIndex, docID));
                path.subList(size, path.size()).clear();
            }
        }

        /**
         *
         */
        public void endElement(String uri, String localName, String qName) {
            path.remove(path.size() - 1);
        }

        /**
         *
         */
        public void characters(char ch[], int start, int length) {
            String text = new String(ch, start, length);

            if (ourIgnoreWhitespace && text.trim().equals("")) {
                return;
            }

            int size = path.size();
            path.addAll(makeTextSymbols(text));
            IWritableNode node = myTrie.add(path);
            node.addReference(new XMLReference(tagIndex, docID));
            path.subList(size, path.size()).clear();
        }
    }

    /**
     * Adds new XML document to the Index Fabric.
     * <p>
     * Method does strictly the same as {@link #addXMLDocument(String, InputStream)}, where
     * the input stream is opened by <tt>docName</tt>.
     *
     * @param   docName                             the document name
     * @see     #addXMLDocument(String, InputStream)
     */
    public void addXMLDocument(String docName) throws IndexException {
        try {
            addXMLDocument(docName, newInputStream(docName));
        } catch (IOException e) {
            throw new IndexException(e);
        }
    }

    /**
     * Adds new XML document to the Index Fabric from specified input stream.
     * <p>
     * If the Index Fabric is unmodifiable, method throws an <tt>UnsupportedOperationException</tt>.
     * Otherwise, as XML document is read from input stream and parsed using SAX {@link #myHandler}.
     *
     * @param   docID                               the document identifier
     * @param   input                               input stream to read document from
     * @throws  UnsupportedOperationException       if the Index Fabric is unmodifiable
     * @see     #addXMLDocument(String)
     */
    public void addXMLDocument(String docID, InputStream input) throws IndexException {
        if (!myTextMap.isModifiable() || !myTagMap.isModifiable() || !myAttrMap.isModifiable()) {
            throw new UnsupportedOperationException("Builder can not add new XML documents");
        }

        try {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setNamespaceAware(true);
            SAXParser parser = factory.newSAXParser();
            XMLReader reader = parser.getXMLReader();

            myHandler.restart(docID);
            reader.setContentHandler(myHandler);

            reader.parse(new InputSource(input));
        } catch (ParserConfigurationException e) {
            throw new IndexException(e);
        } catch (SAXException e) {
            throw new IndexException(e);
        } catch (IOException e) {
            throw new IndexException(e);
        }
    }

    /**
     * Returns the Index Fabric.
     * <p>
     * After constructing IndexFabric, no other addings to Index Fabric are admissible.
     *
     * @return                                      the Index Fabric
     */
    public IndexFabric getIndexFabric() {
        myTrie.setModifiable(false);
        myTextMap.setModifiable(false);
        myTagMap.setModifiable(false);
        myAttrMap.setModifiable(false);

        return new IndexFabric(myTrie, myTextMap, myTagMap, myAttrMap);
    }

    /**
     * Returns current string information about <tt>XMLBuilder</tt>.
     * <p>
     * Mainly can be used in tests.
     *
     * @return                                      a string representation of builder
     */
    public String toString() {
        return "Index Fabric builder of XML\n" +
                "trie size = " + myTrie.size() + "\n" +
                "text map size = " + myTextMap.size() + "\n" +
                "attr map size = " + myAttrMap.size() + "\n" +
                "tag map size = " + myTagMap.size() + "\n" +
                "total map size = " + (myTextMap.size() + myAttrMap.size() + myTagMap.size());
    }
}
